## ----------------------------------------------------------------------------
# Rishu::Remote
# -----------------------------------------------------------------------------
# 里珠・外部のプロセスの受け側.
# -----------------------------------------------------------------------------
# Mastering programed by YAMASHINA Hio
#
# Copyright 2002 YMIRLINK,Inc.
# -----------------------------------------------------------------------------
# $Id: Remote.pm,v 1.14 2002/12/21 09:28:59 hio Exp $
# -----------------------------------------------------------------------------
package Rishu::Remote;
use strict;
use vars qw($INSTANCE);
use vars qw($VERSION);
$VERSION = '1.1.27';

# -----------------------------------------------------------------------------
# Rishu::Remote->instance
#   インスタンスの取得.
#   (初回のみ新しいインスタンスの作成 ＆ 登録)
sub instance
{
  unless( $INSTANCE )
  {
    local($_[0]) = 'Rishu::Remote';
    $INSTANCE = &new;
  }
  $INSTANCE;
}

# -----------------------------------------------------------------------------
# Rishu::Remote->new
sub new
{
  my $pkg = shift;
  my $this = bless{},$pkg;
  my @arg = @_;
  
  $this->{logbits}{system} = 1;
  $this->{logbits}{error}  = 1;
  $this->{logbits}{trace}  = 1;
  $this->{logbits}{recv}   = 1;
  $this->{logbits}{send}   = 1;

  $this->{on_load}    = \&main::on_load;
  $this->{on_request} = \&main::on_request;
  $this->{on_unload}  = \&main::on_unload;
  $this->{on_exit} = sub{};
  
  $this->log('trace','arg: '.join(' ',map{qq/"$_"/}@arg)."\n");
  
  my ($rd,$wd);
  while( @arg )
  {
    my $arg = shift(@arg);
    if( $arg && $arg eq '--port' )
    {
      my $port = shift(@arg);
      use Rishu::Gensym;
      use Socket;
      my $addr = sockaddr_in($port,INADDR_LOOPBACK);
      #$addr = sockaddr_in($port,inet_aton('neige'));
      my $sock = Rishu::Gensym::gensym();
      if( !socket($sock,PF_INET,SOCK_STREAM,getprotobyname('tcp')) )
      {
	$this->log('system',
		   "socket() failed\n[[\n$!\n]]\n"
		   );
	die "socket(), failed : $!";
      }
      binmode($sock);
      if( !connect($sock,$addr) )
      {
	$this->log('system',
		   "connect() failed\n[[\n$!\n]]\n"
		   );
	die "connect(), failed : $!";
      }
      $this->log('system',
		 "connect() ok\n"
		 );
      ($rd,$wd) = ($sock,$sock);
      last;
    }
  }
  select((select($wd),$|=1)[0]);
  
  $rd or binmode(STDIN), $rd=\*STDIN;
  $wd or binmode(STDOUT),$wd=\*STDOUT;
  $this->{rd} = $rd;
  $this->{wd} = $wd;
  
  $this;
}

# -----------------------------------------------------------------------------
# Rishu::Remote->DESTROY
sub DESTROY
{
  my $this = shift;
  
  $this->{logfilename} and close LOG;
}

# -----------------------------------------------------------------------------
# Rishu::Remote->run()
sub run
{
  my $this = shift;
  my $rd = $this->{rd};
  $this->log("system","begin running\n");
  for(;;)
  {
    # コマンド行の捜索.
    $_ = $this->_readLine();
    if( !defined($_) )
    {
      # EOF.
      last;
    }
    s/\r?\n//;
    $this->{logbits}{recv} and $this->log("recv","$_\n");
    if( $_ eq '' )
    {
      # 空行が先行してたらそれは無視.
      next;
    }
    
    # ヘッダ部分の認識
    my $requestLine = $_;
    my ($method,$type,$proto,$ver) = m,^(\S+)\s+(\S+)\s+([^\s/]+)/(\d+\.\d+)$,i;
    
    my $in = '';
    my $contentLength;
    while($_ = $this->_readLine())
    {
      s/\r?\n//;
      $this->{logbits}{recv} and $this->log("recv","$_\n");
      $in .= $_."\r\n";
      if( $_ eq '' )
      {
	# ヘッダの終わりの認識.
	last;
      }
      if( !defined($contentLength) && m/^Content-Length:\s*(\d+)/i )
      {
	$contentLength = $1;
	$this->{logbits}{trace} and $this->log("trace","(content-length: $contentLength)\n");
      }
    }
    $this->{logbits}{trace} and $this->log("trace","(end-of-header)\n");
    if( $contentLength )
    {
      my $body = $this->_readBytes($contentLength);
      if( !defiend($body) || length($body)!=$contentLength )
      {
	my $nRes = defined($body) ? length($body) : 0;
	$this->{logbits}{error} and $this->log("error","expected $contentLength bytes, but read only $nRes bytes.\n");
      }
    }
    if( !$type )
    {
      $this->replyBadRequest("invalid request line: $requestLine");
      next;
    }
    $type = uc($type);
    if( $type eq 'EXIT' )
    { # そのまま終了.
      $this->{logbits}{trace} and $this->log("trace","(call:on_exit)\n");
      return $this->{on_exit}->();
    }elsif( $type eq 'LOAD' )
    {
      $this->{logbits}{trace} and $this->log("trace","(prepare on_load)\n");
      my $dir = join '',grep{s/^Directory: //} split(/\r?\n/,$in);
      if( $dir ne '' )
      {
	$this->{logbits}{trace}
	  and $this->log("trace","(call:on_load) Directory: $dir\n");
        my $res = $this->{on_load}->($dir);
	my $reply;
	if( $res )
	{
	  $reply = "RISHU/1.1 200 OK\r\n";
	  $reply .= "Sender: Rishu/Proxy\r\n";
	}else
	{
	  $reply = "RISHU/1.1 500 Internal Server Error\r\n";
	  $reply .= "Sender: Rishu/Proxy\r\n";
	}
	if( !defined($res) || length($res)==0 )
	{
	  $reply .= "\r\n";
        }else
	{
	  $reply .= "Content-Length: ".length($res)."\r\n";
	  $reply .= "\r\n";
	  $reply .= $res;
	}
        $this->_print($reply);
      }else
      {
	$this->replyBadRequest("LOAD, no Directory: element");
      }
    }elsif( $type eq 'REQUEST' )
    {
      $in =~ s/^Command:\s+//;
      my $res = $this->{on_request}->($in);
      $this->_print("RISHU/1.1 200 OK\r\nOriginal-Response: ".join("\r\n",split(/\r?\n/,$res),"\r\n"));
    }elsif( $type eq 'UNLOAD' )
    {
      my $res = $this->{on_unload}->($in);
      my $reply;
      if( $res )
      {
	$reply = "RISHU/1.1 200 OK\r\n";
	$reply .= "Sender: Rishu/Proxy\r\n";
      }else
      {
	$reply = "RISHU/1.1 500 Internal Server Error\r\n";
	$reply .= "Sender: Rishu/Proxy\r\n";
      }
      if( !defined($res) || length($res)==0 )
      {
	$reply .= "\r\n";
      }else
      {
	$reply .= "Content-Length: ".length($res)."\r\n";
	$reply .= "\r\n";
	$reply .= $res;
      }
      $this->_print($reply);
    }else
    {
      $this->replyBadRequest("type: $type");
    }
  }
  
  $this;
}

# -----------------------------------------------------------------------------
# $remote->replyBadRequest
sub replyBadRequest
{
  my $this = shift;
  my $cause = shift;

  my $reply = "RISHU/1.1 400 Bad Request\r\n";
  $reply .= "Sender: Rishu/Proxy\r\n";
  $reply .= "Rishu-Message: $cause\r\n";
  $reply .= "\r\n";
  $this->_print($reply);
  $this;
}

# -----------------------------------------------------------------------------
# $remote->_print
sub _print
{
  my $this = shift;
  my $sym = $this->{wd};
  my $msg = join('',@_);
  syswrite($sym,$msg);
  $this->{logbits}{send} and $this->log('send',$msg);
  
  $this;
}

# -----------------------------------------------------------------------------
# $remote->_readLine()
sub _readLine
{
  my $this = shift;
  my $rd = $this->{rd};
  if( !exists($this->{rd_buf}) )
  {
    $this->{rd_buf} = '';
  }
  my $nRes;
  for(;;)
  {
    my $pos = index($this->{rd_buf},"\n");
    if( $pos!=-1 )
    {
      my $data = substr($this->{rd_buf},0,$pos+1);
      $this->{rd_buf} = substr($this->{rd_buf},$pos+1);
      return $data;
    }
    $nRes = sysread($rd,$this->{rd_buf},1024,length($this->{rd_buf}));
    if( !$nRes )
    {
      defined($nRes) or $this->log('trace',"sysread failed: $!\n");
      (my $data,$this->{rd_buf}) = ($this->{rd_buf},'');
      return $data;
    }
  }
}

# -----------------------------------------------------------------------------
# $remote->_readBytes()
sub _readBytes
{
  my $this = shift;
  my $required = shift;
  
  my $rd = $this->{rd};
  
  if( $required<0 )
  {
    return '';
  }
  
  if( !exists($this->{rd_buf}) )
  {
    $this->{rd_buf} = '';
  }
  my $nRes;
  while( length($this->{rd_buf})<$required )
  {
    $nRes = sysread($rd,$this->{rd_buf},1024,length($this->{rd_buf}));
    if( !defined($nRes) )
    {
      last;
    }
  }
  my $data = substr($this->{rd_buf},0,$required);
  $this->{rd_buf} = substr($this->{rd_buf},$required);
  $data;
}

# -----------------------------------------------------------------------------
# $remote->log('type','text');
sub log
{
  my $this = shift;
  my $type = shift;
  my $msg = shift;
  
  if( !$this->{logbits}{$type} )
  {
    # ログ対象外はさくっとカット.
    return $this;
  }
  
  if( !$this->{logfilename} )
  {
    # ログファイルのオープン.
    my $fn = $0;
    $fn =~ s/\.pl$/_log.txt/ or $fn .= "_log.txt";  
    $this->{logfilename} = $fn;
    if( -e $fn )
    {
      (my $bak = $fn)=~s/txt$/bak/;
      rename($fn,$bak);
    }
    if( !open(LOG,">>$this->{logfilename}") )
    {
      delete $this->{logfilename};
      return;
    }
    binmode(LOG);
    select((select(LOG),$|=1)[0]);
  }
  
  use Time::HiRes;
  my ($timesec,$usec) = Time::HiRes::gettimeofday();
  #   0    1    2     3     4    5     6     7     8
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)
      = localtime($timesec);
  ++$mon;
  $year += 1900;
  my $time = sprintf("%02d/%02d %02d:%02d:%02d.%06d",
		     $mon,$mday,$hour,$min,$sec,$usec);
  
  my $head = "$time $type> ";
  $msg =~ s/\n(?=.)/\n$head/sg;
  print LOG $head.$msg;

  $this;
}

# -----------------------------------------------------------------------------
# end of package
# -----------------------------------------------------------------------------
1;
